import os
import math
import requests
from .xyzgeneric import *

__TD = 256

def calc_xyz_from_lonlat(lon: float, lat: float, z: int):
    """Calculates XYZ tile indices of the given zoom level at the given longitude and latitude.

    Args:
        lon (float): longitude (degree).
        lat (float): latitude (degree).
        z (int): zoom level (0 or greater).

    Returns:
        (x, y, z) (tupple): tile indices x and y of the zoom level z.

    """
    if z < 0:
        raise ValueError("Zoom level must be 0 or greater.")
    #print(lon, lat, z)
    n = 2.0 ** z
    x = int((lon + 180.0) / 360.0 * n)
    y = int( (1.0 -np.arctanh(np.sin(np.radians(lat))) / np.pi) * n )
    return (x, y, z)

def calc_lonlat_from_xyz(x:int, y:int, z:int):
    """Calculates (lon, lat) degrees of the north west corner of the tile
    """
    n = 2.0 ** z
    lon_deg = x / n * 360.0 - 180.0
    ycorr = np.pi * (1 - y / n)
    lat_rad = np.arcsin(np.tanh(ycorr))
    lat_deg = np.degrees(lat_rad)
    return lon_deg, lat_deg


def calc_bounds_xyz(x:int, y:int, z:int):
    """Calculates the bounds (west, south, east, north) of the tile
    """
    west, north = calc_lonlat_from_xyz(x, y, z)
    east, south = calc_lonlat_from_xyz(x+1, y+1, z)
    return west, south, east, north

class XYZTileFile:
    """Factory class of XYZFiletype instances.

    An instance of the appropriate XYZFiletype class will be returned by judging the extension of the given base string. This class will call XYZHttpTileFile constructor if the base start with "http://".

    Args:
        base (string): base url or file path of xyz file system. This must contain keywords "{x}", "{y}", and "{z}".

    Optional Keyargs:
        loadfunc (method): (optional) user-defined loading function.
        savefunc (method): (optional) user-defined saving function.

    """
    typeclass = {"generic":XYZGeneric}
    def __new__(cls, base: str, type=None, **kwargs):
        if isinstance(base, str):
            if base[0:7] == "http://" or base[0:8] == "https://":
                return XYZHttpTileFile(base, **kwargs)
            if type in cls.typeclass: # try user-specified type
                return cls.typeclass[type](base, **kwargs)
            type = os.path.splitext(base)[-1][1:]
            if type in cls.typeclass:
                return cls.typeclass[type](base, **kwargs)
        return XYZGeneric(base=base, **kwargs)

class XYZHttpTileFile(XYZTileFile):
    """Factory class of XYZHttpFiletype instances.

    An instance of the appropriate XYZHttpFiletype class will be returned by judging the extension of the given base string. This class assumes that the base starts with "http://".

    Args:
        base (string): base url or file path of xyz file system. This must contain keywords "{x}", "{y}", and "{z}".

    Optional Keyargs:
        loadfunc (method): (optional) user-defined loading function.
        savefunc (method): (optional) user-defined saving function.

    """
    typeclass = {"generic":XYZHttpGeneric}
    def __new__(cls, base: str, type=None, **kwargs):
        if isinstance(base, str):
            if base[0:7] != "http://" and base[0:8] != "https://":
                return XYZTileFile(base, **kwargs)
            if type in cls.typeclass:
                return cls.typeclass[type](base, **kwargs)
            type = os.path.splitext(base)[-1][1:]
            if type in cls.typeclass:
                return cls.typeclass[type](base, **kwargs)
        return XYZHttpGeneric(base=base, **kwargs)


for module in os.listdir(os.path.dirname(__file__)):
    if module == '__init__.py' or module == "xyztilefile.py" \
        or module == "xyzgeneric.py" or module[-3:] != '.py':
        continue
    #print(f"Importing {module}")
    exec(f"from .{module[:-3]} import *" )
    XYZTileFile.typeclass.update(xyztiletype)
    XYZHttpTileFile.typeclass.update(xyzhttptype)
del module
del xyztiletype
del xyzhttptype

#print(XYZTileFile.typeclass)



import numpy as np
from .xyzgeneric import *

# Default functions for loading and saving files
_loadfunc = lambda byteio : np.load(byteio)

_savefunc = lambda filename, val : np.save(filename,val)

class XYZNpy(XYZGeneric):
    def __init__(self, base, loadfunc=_loadfunc, savefunc=_savefunc, **kwargs):
        super().__init__(base, loadfunc=loadfunc, savefunc=savefunc, **kwargs)

class XYZHttpNpy(XYZHttpGeneric):
    def __init__(self, base, loadfunc=_loadfunc, **kwargs):
        super().__init__(base, loadfunc=_loadfunc, **kwargs)

# {"type string from the extension" : Class}
xyztiletype = {"npy":XYZNpy}
xyzhttptype = {"npy":XYZHttpNpy}



import io
from .xyzgeneric import *

# Default functions for loading and saving files
_loadfunc = lambda byteio : byteio.read().decode(encoding="utf-8").rstrip()

_savefunc = lambda filename, val : print(val, file=open(filename,"w"))

class XYZTxt(XYZGeneric):
    def __init__(self, base, loadfunc=_loadfunc, savefunc=_savefunc, **kwargs):
        super().__init__(base, loadfunc=loadfunc, savefunc=savefunc, **kwargs)

class XYZHttpTxt(XYZHttpGeneric):
    def __init__(self, base, loadfunc=_loadfunc, **kwargs):
        super().__init__(base, loadfunc=_loadfunc, **kwargs)

# {"type string from the extension" : Class}
xyztiletype = {"txt":XYZTxt}
xyzhttptype = {"txt":XYZHttpTxt}



import pickle
from .xyzgeneric import *

# Default functions for loading and saving files
_loadfunc = lambda byteio : pickle.load(byteio)

_savefunc = lambda filename, val : pickle.dump(val, open(filename,"wb"))

class XYZPickle(XYZGeneric):
    def __init__(self, base, loadfunc=_loadfunc, savefunc=_savefunc, allow_pickle=False, **kwargs):
        if not allow_pickle:
            raise RuntimeError("The pickle module is not secure. Only unpickle data you trust.\nIt is possible to construct malicious pickle data which will execute arbitrary code during unpickling. Never unpickle data that could have come from an untrusted source, or that could have been tampered with.\nWith trusted data source, set allow_pickle to True at the initialization of XYZTileFile.")
        super().__init__(base, loadfunc=loadfunc, savefunc=savefunc, **kwargs)

class XYZHttpPickle(XYZHttpGeneric):
    def __init__(self, base, loadfunc=_loadfunc, **kwargs):
        raise NotImplementedError("Pickle is not allowed on XYZTileFile with HTTP due to security reasons.")


# {"type string from the extension" : Class}
xyztiletype = {"pickle":XYZPickle,"pkl":XYZPickle}
xyzhttptype = {"pickle":XYZHttpPickle,"pkl":XYZHttpPickle}



import os
from .xyztilefile import *

class XYZTileManager:
    """Utilize locally saved cache files with the source

    The source can be local or online.
    Firstly, this class try to load from local cache file.
    If there is no cache file, this class retrieves data from the source.
    The retrieve data is automatically saved to local cache file.

    """
    def __init__(self, srcbase:str, cachebase:str, **kwargs):
        self._cache = {}
        self._src = XYZTileFile(srcbase, cache=self._cache, **kwargs)
        self._lcl = XYZHttpTileFile(cachebase, cache=self._cache, **kwargs)

    def get(self, x:int, y:int, z:int):
        try:
            return self._lcl.get(x,y,z)
        except OSError:
            pass
        res = self._src.get(x,y,z)
        print(self._lcl)
        self._lcl.save(x,y,z)
        return res

    def __repr__(self):
        return f"<{repr(self.__class__)}: {repr(self._src)}, {repr(self._lcl)}>"



from xyztilefile import *
import pprint
from datetime import  datetime

pp = pprint.PrettyPrinter(indent=3)

pp.pprint(calc_xyz_from_lonlat(135.0,35.0,15))
x, y, z = calc_xyz_from_lonlat(135.0,35.0,15)

test= XYZTileFile("./test/{z}/{x}/{y}.png")
pp.pprint(test)

test_http = XYZTileFile("http://localhost:8080/tile/{z}/{x}/{y}.png")
pp.pprint(test_http)
try:
    pp.pprint(test_http.save())
except NotImplementedError as e:
    print(f"NotImplementedError is detected as expected: {e}")
#test_http.get(x,y,z)


test_sample = XYZTileFile("./tile/{z}/{x}/{y}.xyzsample")
pp.pprint(test_sample)
try:
    data = test_sample.get(x,y,z)
except OSError as e:
    print(f"OSError is detected as expected: {e}")

try:
    test_error = XYZTileFile(None)
except TypeError as e:
    print(f"TypeError is detected as expected: {e}")

try:
    test_error = XYZTileFile("http://google.com")
except ValueError as e:
    print(f"ValueError is detected as expected: {e}")


xyz = calc_xyz_from_lonlat

test_txt = XYZTileFile("./tile_sample/{z}/{x}/{y}.txt")
pp.pprint(test_txt)
txt = test_txt.get(*xyz(135.0,35.0,15))
pp.pprint(txt)
pp.pprint(test_txt.set(x,y,z,txt))#+"\nNew line is added"))
pp.pprint(test_txt.save(x,y,z))


test_json = XYZTileFile("./tile_sample/{z}/{x}/{y}.json")
pp.pprint(test_json)
data = test_json.get(*xyz(135.0,35.0,15))
pp.pprint(data)
data.append({"id":3, "txt":f"new line as {datetime.now()}"})
pp.pprint(test_json.set(x-1,y,z,data))
pp.pprint(test_json.save_all())


print()
print("Testing HTTP Text")
test_httptxt = XYZTileFile("https://raw.githubusercontent.com/aDAVISk/xyztilefile/dev/tile_sample/{z}/{x}/{y}.txt")
pp.pprint(test_httptxt)
txt = test_httptxt.get(*xyz(135.0,35.0,15))
pp.pprint(txt)
print(txt)
pp.pprint(test_httptxt)

try:
    txt = test_httptxt.get(*xyz(0,0,15))
except OSError as e:
    print(f"OSError is detected as expected: {e}")

print()
print("Testing HTTP JSON")
test_httpjson = XYZTileFile("https://raw.githubusercontent.com/aDAVISk/xyztilefile/dev/tile_sample/{z}/{x}/{y}.json")
pp.pprint(test_json)
data = test_json.get(*xyz(135.0,35.0,15))
pp.pprint(data)




import geojson
from shapely.geometry import shape
from .xyzgeneric import *

# Default functions for loading and saving files
_loadfunc = lambda byteio: geojson.loads(byteio.read())

_savefunc = lambda filename, val : print(geojson.dumps(val), file=open(filename,"w"))

class XYZGeoJson(XYZGeneric):
    def __init__(self, base, loadfunc=_loadfunc, savefunc=_savefunc, **kwargs):
        super().__init__(base, loadfunc=loadfunc, savefunc=savefunc, **kwargs)

class XYZHttpGeoJson(XYZHttpGeneric):
    def __init__(self, base, loadfunc=_loadfunc, **kwargs):
        super().__init__(base, loadfunc=_loadfunc, **kwargs)


# {"type string from the extension" : Class}
xyztiletype = {"geojson":XYZGeoJson}
xyzhttptype = {"geojson":XYZHttpGeoJson}



import json
from .xyzgeneric import *

# Default functions for loading and saving files
_loadfunc = lambda byteio : json.load(byteio)

_savefunc = lambda filename, val : json.dump(val, open(filename,"w"))

class XYZJson(XYZGeneric):
    def __init__(self, base, loadfunc=_loadfunc, savefunc=_savefunc, **kwargs):
        super().__init__(base, loadfunc=loadfunc, savefunc=savefunc, **kwargs)

class XYZHttpJson(XYZHttpGeneric):
    def __init__(self, base, loadfunc=_loadfunc, **kwargs):
        super().__init__(base, loadfunc=_loadfunc, **kwargs)


# {"type string from the extension" : Class}
xyztiletype = {"json":XYZJson}
xyzhttptype = {"json":XYZHttpJson}



import pprint
import numpy as np
from xyztilefile import *

pp = pprint.PrettyPrinter(indent=3)

x, y, z = calc_xyz_from_lonlat(135.0,35.0,15)
xyz = calc_xyz_from_lonlat

data = {"var":[1,2,3]}
print("Initial")
try:
    test = XYZTileFile("./tile_sample2/{z}/{x}/{y}.pkl")#, allow_pickle=True)
    pp.pprint(test)

    print()
    print("Set data")
    test.set(x,y,z,data)
    pp.pprint(test)

    print()
    print("Save & clear cache")
    test.save(x,y,z)
    test.clc()
    pp.pprint(test)

    print()
    print("Load the saved data")
    data2 = test.get(*xyz(135.0,35.0,15))
    pp.pprint(data2)
    pp.pprint(test)
except RuntimeError as e:
    print(f"RuntimeError is detected as expected: {e}")

print()
print("Testing HTTP npy")
try:
    test_http = XYZTileFile("https://raw.githubusercontent.com/aDAVISk/xyztilefile/dev/tile_sample/{z}/{x}/{y}.pkl")
    pp.pprint(test_http)
except NotImplementedError as e:
    print(f"NotImplementedError is detected as expected: {e}")



import json
from skimage import io as skiio
from .xyzgeneric import *

# Default functions for loading and saving files
_loadfunc = lambda byteio : skiio.imread(byteio)

_savefunc = lambda filename, val : skiio.imsave(filename, val, plugin='imageio')

class XYZPng(XYZGeneric):
    def __init__(self, base, loadfunc=_loadfunc, savefunc=_savefunc, **kwargs):
        super().__init__(base, loadfunc=loadfunc, savefunc=savefunc, **kwargs)

class XYZHttpPng(XYZHttpGeneric):
    def __init__(self, base, loadfunc=_loadfunc, **kwargs):
        super().__init__(base, loadfunc=_loadfunc, **kwargs)


# {"type string from the extension" : Class}
xyztiletype = {"png":XYZPng}
xyzhttptype = {"png":XYZHttpPng}



import warnings
import os
import io
import requests
import copy

# Default functions for loading and saving files
_loadfunc = lambda byteio : byteio.read()

_savefunc = lambda filename, val : print(filename, val)

class XYZGeneric:
    def __init__(self, base: str, loadfunc=_loadfunc, savefunc=_savefunc, cache=None):
        if not isinstance(base, str):
            raise TypeError("base must be string.")
        if not ("{x}" in base and "{y}" in base and "{z}" in base):
            raise ValueError('base must contain "{x}", "{y}", and "{z}".')
        self._base = base
        self._keyfmt = "{z}/{x}/{y}"
        self._loadfunc = loadfunc
        if loadfunc is _loadfunc:
            warnings.warn("XYZGeneric: default loading function is set.")
        self._savefunc = savefunc
        if savefunc is _savefunc:
            warnings.warn("XYZGeneric: default saving function is set.")
        self.clc(cache=cache)

    def __repr__(self):
        # ref: https://ja.pymotw.com/2/pprint/
        return f"<{repr(self.__class__)}: base={repr(self._base)}, cache={repr(self._cache)}>"

    def has(self, x:int, y:int, z:int):
        key = self._keyfmt.format(x=x,y=y,z=z)
        return key in self._base

    def get(self, x:int, y:int, z:int):
        key = self._keyfmt.format(x=x,y=y,z=z)
        if key not in self._base:
            with open(self._base.format(x=x,y=y,z=z), "rb") as ifile:
                self._cache[key] = self._loadfunc(ifile)
        return copy.deepcopy(self._cache[key])

    def set(self, x:int, y:int, z:int, val):
        key = self._keyfmt.format(x=x,y=y,z=z)
        self._cache[key] = val
        return self

    def save(self, x:int, y:int, z:int):
        key = self._keyfmt.format(x=x,y=y,z=z)
        #print(key)
        if key not in self._cache:
            return False
        path = self._base.format(x=x,y=y,z=z)
        dir = os.path.dirname(path)
        print(f"Saving check : {dir} : {os.path.exists(dir)}")
        if not os.path.exists(dir):
            os.makedirs(dir)
        self._savefunc(path, self._cache[key])
        return True

    def set_save(self, x:int, y:int, z:int, val):
        self.set(self, x, y, z, val)
        return self.save(self, x, y, z)

    def save_all(self):
        for key in self._cache:
            dir = os.path.dirname(key)
            if not os.path.exists(dir):
                os.makedirs(dir)
            self._savefunc(key, self._cache[key])
        return True

    def clc(self, cache=None):
        if cache is None:
            cache = {}
        if not isinstance(cache, dict):
            raise TypeError("chache must be dict.")
        if len(cache) > 0:
            warnings.warn("XYZGeneric: non-empty shared chache is set.")
        self._cache = cache

class XYZHttpGeneric(XYZGeneric):
    def __init__(self, base, loadfunc=_loadfunc, **kwargs):
        super().__init__(base, loadfunc=loadfunc, savefunc=None, **kwargs)

    def get(self, x:int, y:int, z:int):
        key = self._keyfmt.format(x=x,y=y,z=z)
        if key not in self._base:
            res = requests.get(self._base.format(x=x,y=y,z=z))
            if res.status_code != 200:
                res.raise_for_status()
                #raise OSError(f"Failed to fetch {key}")
            self._cache[key] = self._loadfunc(io.BytesIO(res.content))
        #copy.deepcopy した方がいい？
        return copy.deepcopy(self._cache[key])


    def set(self, *args, **kwargs):
        raise NotImplementedError("set is nulled for http tile class.")

    def save(self, *args, **kwargs):
        raise NotImplementedError("save is nulled for http tile class.")

    def save_all(self, *args, **kwargs):
        raise NotImplementedError("save_all is nulled for http tile class.")


# {"type string from the extension" : Class}
# These variables are not used for Generic classes
#xyztiletype = {"xyzgeneric":XYZGeneric}
#xyzhttptype = {"xyzgeneric":XYZHttpGeneric}



from .xyzgeneric import *

# Default functions for loading and saving files
_loadfunc = lambda byteio : byteio.read()

_savefunc = lambda filename, val : print(filename, val)

class XYZSample2(XYZGeneric):
    def __init__(self, base, loadfunc=_loadfunc, savefunc=_savefunc, **kwargs):
        super().__init__(base, loadfunc=loadfunc, savefunc=savefunc, **kwargs)

class XYZHttpSample2(XYZHttpGeneric):
    def __init__(self, base, loadfunc=_loadfunc, **kwargs):
        super().__init__(base, loadfunc=_loadfunc, **kwargs)


# {"type string from the extension" : Class}
xyztiletype = {"xyzsample2":XYZSample2}
xyzhttptype = {"xyzsample2":XYZHttpSample2}



from xyztilefile import *
import pprint
from datetime import  datetime

pp = pprint.PrettyPrinter(indent=3)

pp.pprint(calc_xyz_from_lonlat(135.0,35.0,15))
xyz = calc_xyz_from_lonlat

print()
print("Testing Tile Manager")
test_json = XYZTileManager("https://raw.githubusercontent.com/aDAVISk/xyztilefile/dev/tile_sample/{z}/{x}/{y}.json", "./tile_sample2/{z}/{x}/{y}.json")
pp.pprint(test_json)
data = test_json.get(*xyz(135.0,35.0,15))
pp.pprint(data)
pp.pprint(test_json)




# ref: https://qiita.com/studio_haneya/items/9aad8f9ede11e58b41a8
# ref: https://qiita.com/shonansurvivors/items/0fbcbfde129f2d26301c

from setuptools import setup, find_packages
import os

# long_description(後述)に、GitHub用のREADME.mdを指定
with open(os.path.join(os.path.dirname(__file__), 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='xyztilefile',  # パッケージ名（pip listで表示される）
    version="0.3.3",  # バージョン
    description="manage files in xyz tile format",  # 説明
    author='Akito Davis Kawamura',  # 作者名
    license='MIT',  # ライセンス
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    install_requires=['numpy', 'scikit-image','geojson'],
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords='xyztilefile xyz-tile xyztile',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.9',
    ],
)



import pprint
import matplotlib.pyplot as plt
import numpy as np
from xyztilefile import *

pp = pprint.PrettyPrinter(indent=3)

x, y, z = calc_xyz_from_lonlat(135.0,35.0,15)
xyz = calc_xyz_from_lonlat

print("Initial")
test = XYZTileFile("./tile_sample/{z}/{x}/{y}.png")
pp.pprint(test)

print()
print("Load the saved image")
img = test.get(*xyz(135.0,35.0,15))
pp.pprint(img)
pp.pprint(test)
print(np.shape(img))
fig, ax = plt.subplots()
ax.imshow(img)
ax.set_title("loaded image")
plt.show()

print()
print("Set array")
img[0:100,-101:-1,:] = [np.random.randint(255), np.random.randint(255), np.random.randint(255), 255]
test.set(x,y,z,img)
pp.pprint(test)

fig, ax = plt.subplots()
ax.imshow(img)
ax.set_title("set image")
plt.show()

print()
print("Save & clear cache")
test.save(x,y,z)
test.clc()
pp.pprint(test)

print()
print("Load the saved image")
img = test.get(*xyz(135.0,35.0,15))
pp.pprint(img)
pp.pprint(test)

fig, ax = plt.subplots()
ax.imshow(img)
ax.set_title("saved & reloaded image")
plt.show()

print()
print("Testing HTTP PNG")
test_http = XYZTileFile("https://raw.githubusercontent.com/aDAVISk/xyztilefile/dev/tile_sample/{z}/{x}/{y}.png")
img = test_http.get(*xyz(135.0,35.0,15))
pp.pprint(img)
pp.pprint(test_http)

fig, ax = plt.subplots()
ax.imshow(img)
ax.set_title("reloaded online image")
plt.show()



import pprint
import numpy as np
from xyztilefile import *

pp = pprint.PrettyPrinter(indent=3)

x, y, z = calc_xyz_from_lonlat(135.0,35.0,15)
xyz = calc_xyz_from_lonlat

ar = np.arange(12).reshape((4,3))
print("Initial")
test = XYZTileFile("./tile_sample/{z}/{x}/{y}.npy")
pp.pprint(test)

print()
print("Set array")
test.set(x,y,z,ar)
pp.pprint(test)

print()
print("Save & clear cache")
test.save(x,y,z)
test.clc()
pp.pprint(test)

print()
print("Load the saved array")
ar2 = test.get(*xyz(135.0,35.0,15))
pp.pprint(ar2)
pp.pprint(test)

print()
print("Testing HTTP npy")
test_http = XYZTileFile("https://raw.githubusercontent.com/aDAVISk/xyztilefile/dev/tile_sample/{z}/{x}/{y}.npy")
ar3 = test_http.get(*xyz(135.0,35.0,15))
pp.pprint(ar3)
#ar3[0,0] = 100
pp.pprint(test_http)
